pub type Scalar = f64;

pub type Vec2 = nalgebra::Vector2<Scalar>;
pub type Point2 = nalgebra::Point2<Scalar>;

pub type Rot2 = nalgebra::UnitComplex<Scalar>;

pub trait Region<T> {
    fn full() -> Self;
    fn singleton(value: T) -> Self;

    fn nearest(&self, value: &T) -> Option<T>;
    fn contains(&self, value: &T) -> bool;
}

#[derive(Clone, Debug)]
pub enum Region1 {
    Empty,
    Singleton(Scalar),
    Range(Scalar, Scalar),
    Union(Box<Region1>, Box<Region1>),
    Full,
}

impl Region<Scalar> for Region1 {
    fn full() -> Self {
        Region1::Full
    }

    fn singleton(value: Scalar) -> Self {
        Region1::Singleton(value)
    }

    fn contains(&self, n: &Scalar) -> bool {
        use Region1::*;
        match self {
            Empty => false,
            Singleton(n1) => relative_eq!(n1, n),
            Range(l, u) => *l <= *n && *n <= *u,
            Union(r1, r2) => r1.contains(n) || r2.contains(n),
            Full => true,
        }
    }

    fn nearest(&self, s: &Scalar) -> Option<Scalar> {
        use Region1::*;
        match self {
            Empty => None,
            Full => Some(*s),
            Singleton(n) => Some(*n),
            Range(l, u) => match (l < s, s < u) {
                (true, true) => Some(*s),
                (true, false) => Some(*u),
                (false, true) => Some(*l),
                _ => None,
            },
            Union(r1, r2) => {
                let distance = |a: Scalar, b: Scalar| (a - b).abs();
                match (r1.nearest(s), r2.nearest(s)) {
                    (None, None) => None,
                    (Some(n), None) | (None, Some(n)) => Some(n),
                    (Some(n1), Some(n2)) => Some({
                        if distance(*s, n1) <= distance(*s, n2) {
                            n1
                        } else {
                            n2
                        }
                    }),
                }
            }
        }
    }
}

// line starting at start, point at angle dir, with range extent
// ie. start + (cos dir, sin dir) * t for t in extent
#[derive(Clone, Debug)]
pub struct Line2 {
    start: Point2,
    dir: Rot2,
    extent: Region1,
}

impl Line2 {
    pub fn new(start: Point2, dir: Rot2, extent: Region1) -> Self {
        Self { start, dir, extent }
    }

    pub fn evaluate(&self, t: Scalar) -> Point2 {
        self.start + self.dir * Vec2::new(t, 0.)
    }

    pub fn nearest(&self, p: &Point2) -> Point2 {
        // rotate angle 90 degrees
        let perp_dir = self.dir * Rot2::from_cos_sin_unchecked(0., 1.);
        let perp = Line2::new(*p, perp_dir, Region1::Full);
        if let Region2::Singleton(np) = self.intersect(&perp) {
            np
        } else {
            panic!("Line2::nearest not found!");
        }
    }

    pub fn intersect(&self, other: &Line2) -> Region2 {
        // if the two lines are parallel...
        let dirs = self.dir / other.dir;
        if relative_eq!(dirs.sin_angle(), 0.) {
            let starts = self.dir.to_rotation_matrix().inverse() * (other.start - self.start);
            return if relative_eq!(starts.y, 0.) {
                // and they are colinear
                Region2::Line(self.clone())
            } else {
                // they are parallel and never intersect
                Region2::Empty
            };
        }
        // TODO: respect extent
        let (a, b) = (self, other);
        let (a_0, a_v, b_0, b_v) = (a.start, a.dir, b.start, b.dir);
        let (a_c, a_s, b_c, b_s) = (
            a_v.cos_angle(),
            a_v.sin_angle(),
            b_v.cos_angle(),
            b_v.sin_angle(),
        );
        let t_b = (a_0.x * a_s - a_0.y * a_c + a_0.x * a_s + b_0.y * a_c) / (a_s * b_c - a_c * b_s);
        Region2::Singleton(b.evaluate(t_b))
    }
}

#[derive(Clone, Debug)]
pub enum Region2 {
    Empty,
    // single point at 0
    Singleton(Point2),
    Line(Line2),
    #[allow(dead_code)]
    Union(Box<Region2>, Box<Region2>),
    Full,
}

impl Region<Point2> for Region2 {
    fn full() -> Self {
        Region2::Full
    }

    fn singleton(value: Point2) -> Self {
        Region2::Singleton(value)
    }

    fn contains(&self, p: &Point2) -> bool {
        self.nearest(p).map_or(false, |n| relative_eq!(n, p))
    }

    fn nearest(&self, p: &Point2) -> Option<Point2> {
        use Region2::*;
        match self {
            Empty => None,
            Full => Some(*p),
            Singleton(n) => Some(*n),
            Line(line) => Some(line.nearest(p)),
            Union(r1, r2) => {
                use nalgebra::distance;
                match (r1.nearest(p), r2.nearest(p)) {
                    (None, None) => None,
                    (Some(n), None) | (None, Some(n)) => Some(n),
                    (Some(n1), Some(n2)) => Some({
                        if distance(p, &n1) <= distance(p, &n2) {
                            n1
                        } else {
                            n2
                        }
                    }),
                }
            }
        }
    }
}

impl Region2 {
    pub fn union(r1: Region2, r2: Region2) -> Region2 {
        use Region2::*;
        match (r1, r2) {
            (Empty, r) | (r, Empty) => r,
            (Full, _) | (_, Full) => Full,
            (r1, r2) => Union(Box::new(r1), Box::new(r2)),
        }
    }

    pub fn intersect(&self, other: &Region2) -> Region2 {
        use Region2::*;
        match (self, other) {
            (Empty, _) | (_, Empty) => Empty,
            (Full, r) | (r, Full) => r.clone(),
            (Singleton(n1), Singleton(n2)) => {
                if n1 == n2 {
                    Singleton(*n1)
                } else {
                    Empty
                }
            }
            (Singleton(n), o) | (o, Singleton(n)) => {
                if o.contains(n) {
                    Singleton(*n)
                } else {
                    Empty
                }
            }
            (Line(l1), Line(l2)) => l1.intersect(l2),
            (Union(un1, un2), o) | (o, Union(un1, un2)) => {
                Self::union(un1.intersect(o), un2.intersect(o))
            }
        }
    }
}

mod solve {

    use std::collections::BTreeSet;
    use std::fmt;
    use std::iter::FromIterator;

    use crate::math::Scalar;

    // an unknown variable with an id
    #[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
    struct Unknown(i64);

    type UnknownSet = BTreeSet<Unknown>;

    trait Unknowns {
        fn unknowns(&self) -> UnknownSet;
        fn has_unknowns(&self) -> bool;
        fn has_unknown(&self, u: Unknown) -> bool;
    }

    impl Unknowns for Scalar {
        fn unknowns(&self) -> UnknownSet {
            UnknownSet::new()
        }
        fn has_unknowns(&self) -> bool {
            false
        }
        fn has_unknown(&self, _: Unknown) -> bool {
            false
        }
    }

    impl Unknowns for Unknown {
        fn unknowns(&self) -> UnknownSet {
            FromIterator::from_iter(Some(*self))
        }
        fn has_unknowns(&self) -> bool {
            true
        }
        fn has_unknown(&self, u: Unknown) -> bool {
            *self == u
        }
    }

    impl fmt::Display for Unknown {
        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
            write!(f, "u{}", self.0)
        }
    }

    #[derive(Clone, Debug, PartialEq)]
    enum Expr {
        Unkn(Unknown),
        Const(Scalar),
        Plus(Box<Expr>, Box<Expr>),
        Neg(Box<Expr>),
        Times(Box<Expr>, Box<Expr>),
        Inv(Box<Expr>),
    }

    impl Unknowns for Expr {
        fn unknowns(&self) -> UnknownSet {
            use Expr::*;
            match self {
                Unkn(u) => u.unknowns(),
                Const(_) => UnknownSet::default(),
                Plus(l, r) | Times(l, r) => l.unknowns().union(&r.unknowns()).cloned().collect(),
                Neg(e) | Inv(e) => e.unknowns(),
            }
        }
        fn has_unknowns(&self) -> bool {
            use Expr::*;
            match self {
                Unkn(u) => u.has_unknowns(),
                Const(_) => false,
                Plus(l, r) | Times(l, r) => l.has_unknowns() || r.has_unknowns(),
                Neg(e) | Inv(e) => e.has_unknowns(),
            }
        }
        fn has_unknown(&self, u: Unknown) -> bool {
            use Expr::*;
            match self {
                Unkn(u1) => u1.has_unknown(u),
                Const(_) => false,
                Plus(l, r) | Times(l, r) => l.has_unknown(u) || r.has_unknown(u),
                Neg(e) | Inv(e) => e.has_unknown(u),
            }
        }
    }

    impl Expr {
        fn new_plus(e1: Expr, e2: Expr) -> Expr {
            Expr::Plus(Box::new(e1), Box::new(e2))
        }
        fn new_times(e1: Expr, e2: Expr) -> Expr {
            Expr::Times(Box::new(e1), Box::new(e2))
        }
        fn new_neg(e1: Expr) -> Expr {
            Expr::Neg(Box::new(e1))
        }
        fn new_inv(e1: Expr) -> Expr {
            Expr::Inv(Box::new(e1))
        }
        fn new_minus(e1: Expr, e2: Expr) -> Expr {
            Expr::Plus(Box::new(e1), Box::new(Expr::new_neg(e2)))
        }
        fn new_div(e1: Expr, e2: Expr) -> Expr {
            Expr::Times(Box::new(e1), Box::new(Expr::new_inv(e2)))
        }

        fn is_zero(&self) -> bool {
            use Expr::*;
            match self {
                Const(c) => relative_eq!(*c, 0.),
                _ => false,
            }
        }

        fn is_one(&self) -> bool {
            use Expr::*;
            match self {
                Const(c) => relative_eq!(*c, 1.),
                _ => false,
            }
        }

        fn simplify(self) -> Expr {
            use Expr::*;
            match self {
                Plus(l, r) => match (l.simplify(), r.simplify()) {
                    (Const(lc), Const(rc)) => Const(lc + rc),
                    (Const(c), ref o) | (ref o, Const(c)) if relative_eq!(c, 0.) => o.clone(),
                    (Times(l1, l2), Times(r1, r2)) => {
                        if l2 == r2 {
                            Expr::new_times(Expr::Plus(l1, r1), *l2).simplify()
                        } else if l1 == r1 {
                            Expr::new_times(Expr::Plus(l2, r2), *l1).simplify()
                        } else if l1 == r2 {
                            Expr::new_times(Expr::Plus(l2, r1), *l1).simplify()
                        } else if l2 == r1 {
                            Expr::new_times(Expr::Plus(l1, r2), *l2).simplify()
                        } else {
                            Expr::new_plus(Times(l1, l2), Times(r1, r2))
                        }
                    }
                    (l, r) => Self::new_plus(l, r),
                },
                Times(l, r) => match (l.simplify(), r.simplify()) {
                    (Const(lc), Const(rc)) => Const(lc * rc),
                    (Const(c), ref o) | (ref o, Const(c)) if relative_eq!(c, 1.) => o.clone(),
                    (Inv(ref den), ref num) | (ref num, Inv(ref den)) if *num == **den => Const(1.),
                    (l, r) => Self::new_times(l, r),
                },
                Neg(v) => match v.simplify() {
                    Const(c) => Const(-c),
                    Neg(v) => *v,
                    e => Self::new_times(Const(-1.), e),
                },
                Inv(v) => match v.simplify() {
                    Const(c) => Const(1. / c),
                    Inv(v) => *v,
                    e => Self::new_inv(e),
                },
                e => e,
            }
        }

        fn distrubute(self) -> Expr {
            use Expr::*;
            match self {
                Plus(l, r) => Expr::new_plus(l.distrubute(), r.distrubute()),
                Times(l, r) => match (*l, *r) {
                    (Plus(l, r), o) | (o, Plus(l, r)) => Expr::new_plus(
                        Expr::Times(Box::new(o.clone()), l),
                        Expr::Times(Box::new(o), r),
                    )
                    .distrubute(),
                    (l, r) => Expr::new_times(l, r),
                },
                Neg(v) => match *v {
                    Plus(l, r) => Expr::new_plus(Neg(l).distrubute(), Neg(r).distrubute()),
                    Times(l, r) => Expr::new_times(Neg(l).distrubute(), *r),
                    Neg(v) => v.distrubute(),
                    Inv(v) => Expr::new_inv(Neg(v).distrubute()),
                    e => Expr::new_neg(e),
                },
                Inv(v) => match *v {
                    Plus(l, r) => Expr::new_plus(Inv(l).distrubute(), Inv(r).distrubute()),
                    Times(l, r) => Expr::new_times(Inv(l).distrubute(), Inv(r).distrubute()),
                    Inv(v) => v.distrubute(),
                    e => Expr::new_inv(e),
                },
                e => e,
            }
        }

        fn reduce(self, for_u: Unknown) -> Expr {
            use Expr::*;
            match self {
                Plus(l, r) => match (l.reduce(for_u), r.reduce(for_u)) {
                    (Const(lc), Const(rc)) => Const(lc + rc),
                    (l, r) => Self::new_plus(l, r),
                },
                Times(l, r) => match (l.reduce(for_u), r.reduce(for_u)) {
                    (Const(lc), Const(rc)) => Const(lc * rc),
                    (l, r) => Self::new_times(l, r),
                },
                Neg(v) => match v.reduce(for_u) {
                    Unkn(u) if u == for_u => Expr::new_times(Const(-1.), Unkn(u)),
                    e => Self::new_neg(e),
                },
                Inv(v) => match v.reduce(for_u) {
                    e => Self::new_inv(e),
                },
                e => e,
            }
        }
    }

    impl fmt::Display for Expr {
        fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
            use Expr::*;
            match self {
                Unkn(u) => write!(f, "{}", u),
                Const(c) => write!(f, "{}", c),
                Plus(l, r) => write!(f, "({}) + ({})", l, r),
                Times(l, r) => write!(f, "({}) * ({})", l, r),
                Neg(e) => write!(f, "-({})", e),
                Inv(e) => write!(f, "1 / ({})", e),
            }
        }
    }

    #[derive(Clone, Debug, PartialEq)]
    struct Eqn(Expr, Expr);

    impl Unknowns for Eqn {
        fn unknowns(&self) -> UnknownSet {
            self.0
                .unknowns()
                .union(&self.1.unknowns())
                .cloned()
                .collect()
        }
        fn has_unknowns(&self) -> bool {
            self.0.has_unknowns() || self.1.has_unknowns()
        }
        fn has_unknown(&self, u: Unknown) -> bool {
            self.0.has_unknown(u) || self.1.has_unknown(u)
        }
    }

    fn ord_by_unkn(a: Expr, b: Expr, u: Unknown) -> Option<(Expr, Expr)> {
        if a.has_unknown(u) {
            Some((a, b))
        } else if b.has_unknown(u) {
            Some((b, a))
        } else {
            None
        }
    }

    impl Eqn {
        fn solve(&self, for_u: Unknown) -> Option<Expr> {
            use Expr::*;
            if !self.has_unknown(for_u) {
                return None;
            }
            let (l, r) = (self.0.clone().simplify(), self.1.clone().simplify());
            let (mut l, mut r) = ord_by_unkn(l, r, for_u)?;
            loop {
                let (new_l, new_r): (Expr, Expr) = match l {
                    Unkn(u) => return if u == for_u { Some(r.simplify()) } else { None },
                    Plus(a, b) => {
                        let (a, b) = ord_by_unkn(*a, *b, for_u)?;
                        (a, Expr::new_minus(r, b))
                    }
                    Times(a, b) => {
                        let (a, b) = ord_by_unkn(*a, *b, for_u)?;
                        (a, Expr::new_div(r, b))
                    }
                    Neg(v) => (*v, Expr::new_neg(r)),
                    Inv(v) => (*v, Expr::new_inv(r)),
                    Const(_) => return None,
                };
                l = new_l;
                r = new_r;
            }
        }
    }

    #[derive(Clone, Debug, PartialEq)]
    struct Eqns(Vec<Eqn>);

    impl Unknowns for Eqns {
        fn unknowns(&self) -> UnknownSet {
            self.0.iter().flat_map(|eqn: &Eqn| eqn.unknowns()).collect()
        }
        fn has_unknowns(&self) -> bool {
            self.0.iter().any(|eqn: &Eqn| eqn.has_unknowns())
        }
        fn has_unknown(&self, u: Unknown) -> bool {
            self.0.iter().any(|eqn: &Eqn| eqn.has_unknown(u))
        }
    }

    #[cfg(test)]
    mod tests {
        use super::*;

        #[test]
        fn test_unknowns() {
            let u1 = Unknown(1);
            let u2 = Unknown(2);
            let u3 = Unknown(3);
            assert!(u1.has_unknowns());
            assert!(u2.has_unknowns());
            assert!(u1.has_unknown(u1));
            assert!(!u1.has_unknown(u2));
            assert!(u1.unknowns().contains(&u1));
            assert!(!u2.unknowns().contains(&u1));

            let e1 = Expr::new_minus(Expr::Unkn(u1), Expr::Unkn(u2));
            assert!(e1.has_unknowns());
            assert!(e1.has_unknown(u1));
            assert!(e1.has_unknown(u2));
            assert!(!e1.has_unknown(u3));
            assert!(e1.unknowns().len() == 2);
        }

        fn const_expr(e: Expr) -> Option<Scalar> {
            match e {
                Expr::Const(c) => Some(c),
                _ => None,
            }
        }

        #[test]
        fn test_solve() {
            use Expr::*;
            let u1 = Unknown(1);
            let e1 = Unkn(u1);
            let e2 = Const(1.);

            let eqn = Eqn(e1.clone(), e2.clone());
            assert_eq!(eqn.solve(u1), Some(Const(1.)));
            let eqn = Eqn(e2.clone(), e1.clone());
            assert_eq!(eqn.solve(u1), Some(Const(1.)));

            let e3 = Expr::new_plus(Const(1.), Const(1.));
            let eqn = Eqn(e1.clone(), e3.clone());
            assert_eq!(eqn.solve(u1), Some(Const(2.)));
            let e3 = Expr::new_minus(Const(1.), Const(1.));
            let eqn = Eqn(e1.clone(), e3.clone());
            assert_eq!(eqn.solve(u1), Some(Const(0.)));

            let e1 = Expr::new_div(Const(2.), Expr::new_minus(Const(1.), Const(4.)));
            let e2 = Expr::new_minus(Const(1.), Unkn(u1));
            let eqn = Eqn(e1, e2);
            let e = eqn.solve(u1).unwrap();
            assert!(const_expr(e.clone()).is_some());
            assert!(relative_eq!(const_expr(e.clone()).unwrap(), 5. / 3.));

            let e1 = Expr::new_times(Const(2.), Expr::new_minus(Const(1.), Const(4.)));
            let e2 = Expr::new_minus(Expr::new_times(Unkn(u1), Const(2.)), Unkn(u1));
            println!(
                "e1==e2: {}=={} => {}=={}",
                e1,
                e2,
                e1.clone().simplify(),
                e2.clone().simplify()
            );
            println!(
                "e1==e2: {}=={} => {}=={}",
                e1,
                e2,
                e1.clone().distrubute(),
                e2.clone().distrubute()
            );
            let eqn = Eqn(e1, e2);
            let e = eqn.solve(u1).unwrap();
            assert!(const_expr(e.clone()).is_some());
            assert!(relative_eq!(const_expr(e.clone()).unwrap(), -6.));
        }

    }
}
